/***********************/
/* losenim.c           */
/* for LoseNIM 1.00    */
/* by Adam M. Costello */
/***********************/

/* This is ANSI C code. */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "nimutils.h"
#include "errmsg.h"


const char * const instructions =
   "Welcome to LoseNIM 1.00 by Adam M. Costello <amc@cs.wustl.edu>.\n"
   "\n"
   "The generalized game of NIM starts with some piles of stones.  Two\n"
   "players take turns choosing a non-empty pile and removing some non-zero\n"
   "number of stones from that pile.  The player who removes the last stone\n"
   "loses.  In this game, you will decide the initial configuration of the\n"
   "stones, and I will decide who goes first.  I will win.\n";


nim_uint getnumstones(errmsg_t errmsg)

/* Reads a nim_uint from stdin, and     */
/* returns the nim_uint, or 0 on error. */
{
  nim_uint numstones;
  int n;

  errmsg[0] = '\0';

  for(;;) {
    n = scanf(NIM_UINT_SCANF_SPEC, &numstones);
    if (n == 1) break;
    if (n == EOF) {
      strcpy(errmsg, "Unexpected end-of-file.\n");
      return 0;
    }
    fprintf(stderr,
            "Skipping garbage input character: %c\n", getchar());
    fflush(stderr);
  }

  return numstones;
}


nimstate askconfig(errmsg_t errmsg)

/* Returns a new nimstate with a configuration decided */
/* interactively by the user, or nostate on failure.   */
{
  nimstate state = nostate;
  nim_uint numstones;

  state = newstate(errmsg);
  if (*errmsg) goto askconfig_err;

  puts(
   "Enter a 0-terminated, whitespace-separated list of positive integers\n"
   "indicating the number of stones initially in each pile."
  );
  fflush(stdout);

  for (;;) {
    numstones = getnumstones(errmsg);
    if (*errmsg) goto askconfig_err;
    if (numstones == 0) break;
    if (newpile(state,numstones,errmsg) == nopile) goto askconfig_err;
  }

  return state;

askconfig_err:

  if (state != nostate) freestate(state);
  return nostate;
}


void printconfig(nimstate state)

/* Prints a description of state to stdout. */
{
  nimpile pile;

  fputs("\nThe configuration is now:", stdout);

  for (pile = firstpile(state);  pile != nopile;  pile = nextpile(pile))
    printf(" %d", pilesize(pile));

  putchar('\n');
}


void domyturn(nimstate state, errmsg_t errmsg)

/* Uses a winning strategy to decide how many stones   */
/* to remove from which pile, and performs the action. */
{
  nim_uint bitstochange, numstones, desired, numtotake;
  nimpile pile;

  errmsg[0] = '\0';

  printconfig(state);

  bitstochange = nimxor(state) ^ (numpluralpiles(state) <= 1);

  if (bitstochange == 0) {
    sprintf(errmsg,impossibility,2);
    goto domyturn_err;
  }

  pile = firstpile(state);
  for (;;) {
    if (pile == nopile) {
      sprintf(errmsg,impossibility,3);
      goto domyturn_err;
    }
    numstones = pilesize(pile);
    desired = numstones ^ bitstochange;
    if (desired <= numstones) break;
    pile = nextpile(pile);
  }

  numtotake = numstones - desired;

  printf("I take %d stone%s from a pile of %d.\n",
         numtotake, (numtotake > 1) ? "s" : "",  numstones);

  (void) diminishpile(pile,numtotake);

domyturn_err:

  return;
}


void doyourturn(nimstate state, errmsg_t errmsg)

/* Asks the user how many stones to remove   */
/* from which pile, and performs the action. */
{
  nim_uint numstones, numtotake;
  nimpile pile;

  for (;;) {
    printconfig(state);

    fputs("From which pile do you wish to remove stones? ", stdout);
    fflush(stdout);

    numstones = getnumstones(errmsg);
    if (*errmsg) goto doyourturn_err;

    pile = findpile(state,numstones);

    if (pile == nopile) {
      printf("There is no pile containing %d stones.  Please choose again.\n",
             numstones);
      continue;
    }

    if (numstones != 0) break;

    puts("You may not choose an empty pile.  Please choose again.");
  }

  for (;;) {
    fputs("How many stones do you wish to remove? ", stdout);
    fflush(stdout);

    numtotake = getnumstones(errmsg);
    if (*errmsg) goto doyourturn_err;

    if (numtotake == 0) {
      puts("You must take some stones.");
      continue;
    }

    if (numtotake <= numstones) break;

    printf("You may not take more stones than are in the pile (%d).\n",
           numstones);
  }

  (void) diminishpile(pile,numtotake);

doyourturn_err:

  return;
}


main()
{
  nimstate state = nostate;
  int myturn;
  errmsg_t errmsg;

  puts(instructions);

  state = askconfig(errmsg);
  if (*errmsg) goto main_err;

  if (totalstones(state) == 0) {
    puts("We cannot play without any stones.  Bye.");
    exit(EXIT_SUCCESS);
  }

  myturn = (nimxor(state) != 0) ^ (numpluralpiles(state) == 0);
             
  fputs(myturn ? "I" : "You", stdout);
  puts(" will go first.");

  do
    if (myturn) {
      domyturn(state,errmsg);
      if (*errmsg) goto main_err;
      myturn = 0;
    }
    else {
      doyourturn(state,errmsg);
      if (*errmsg) goto main_err;
      myturn = 1;
    }
  while (totalstones(state) != 0);

  printconfig(state);

  if (myturn) puts("You took the last stone.  I win.");
  else {
    puts("I took the last stone.  You win.");
    sprintf(errmsg,impossibility,1);
    goto main_err;
  }

  freestate(state);

  exit(EXIT_SUCCESS);

main_err:

  if (state != nostate) freestate(state);
  fputs(errmsg,stderr);
  exit(EXIT_FAILURE);
}
